from posix.types cimport off_t

from libeasel.alphabet cimport ESL_ALPHABET
from libhmmer cimport p7_NEVPARAM, p7_NCUTOFFS, p7_NOFFSETS
from libhmmer.p7_hmm cimport p7_MAXABET


cdef extern from "hmmer.h" nogil:

    const size_t p7P_NXSTATES
    cdef enum p7p_xstates_e:
        p7P_E = 0
        p7P_N = 1
        p7P_J = 2
        p7P_C = 3

    const size_t p7P_NXTRANS
    cdef enum p7p_xtransitions_e:
        p7P_LOOP = 0
        p7P_MOVE = 1

    const size_t p7P_NTRANS
    cdef enum p7p_tsc_e:
        p7P_MM = 0
        p7P_IM = 1
        p7P_DM = 2
        p7P_BM = 3
        p7P_MD = 4
        p7P_DD = 5
        p7P_MI = 6
        p7P_II = 7

    const size_t p7P_NR
    cdef enum p7p_rsc_e:
        p7P_MSC = 0
        p7P_ISC = 1

    cdef enum:
        p7_NO_MODE   = 0
        p7_LOCAL     = 1
        p7_GLOCAL    = 2
        p7_UNILOCAL  = 3
        p7_UNIGLOCAL = 4

    ctypedef p7_profile_s P7_PROFILE
    cdef struct p7_profile_s:
        float* tsc
        float** rsc
        float[p7P_NXSTATES][p7P_NXTRANS] xsc

        int mode
        int L
        int allocM
        int M
        int max_length
        float nj

        char* name
        char* acc
        char* desc
        char* rf
        char* mm
        char* cs
        char* consensus
        float[p7_NEVPARAM] evparam
        float[p7_NCUTOFFS] cutoff
        float[p7_MAXABET] float

        off_t[p7_NOFFSETS] offs
        off_t roff
        off_t eoff

        const ESL_ALPHABET* abc


    P7_PROFILE *p7_profile_Create(int M, const ESL_ALPHABET *abc)
    P7_PROFILE *p7_profile_Clone(const P7_PROFILE *gm)
    int         p7_profile_Copy(const P7_PROFILE *src, P7_PROFILE *dst)
    int         p7_profile_SetNullEmissions(P7_PROFILE *gm)
    int         p7_profile_Reuse(P7_PROFILE *gm)
    size_t      p7_profile_Sizeof(P7_PROFILE *gm)
    void        p7_profile_Destroy(P7_PROFILE *gm)
    int         p7_profile_IsLocal(const P7_PROFILE *gm)
    int         p7_profile_IsMultihit(const P7_PROFILE *gm)
    int         p7_profile_GetT(const P7_PROFILE *gm, char st1, int k1, char st2, int k2, float *ret_tsc);
    int         p7_profile_Validate(const P7_PROFILE *gm, char *errbuf, float tol)
    int         p7_profile_Compare(P7_PROFILE *gm1, P7_PROFILE *gm2, float tol)



cdef extern from "esl_heap.h" nogil:

    cdef size_t eslHEAP_INITALLOC
    cdef int eslHEAP_MIN
    cdef int eslHEAP_MAX

    ctypedef esl_heap_s ESL_HEAP
    cdef struct esl_heap_s:
        int* idata
        int n
        int nalloc
        int maxormin

    ESL_HEAP *esl_heap_ICreate   (int maxormin);
    int       esl_heap_GetCount  (ESL_HEAP *hp);
    int       esl_heap_IGetTopVal(ESL_HEAP *hp);
    int       esl_heap_Reuse     (ESL_HEAP *hp);
    void      esl_heap_Destroy   (ESL_HEAP *hp);

    int       esl_heap_IInsert(ESL_HEAP *hp, int val);

    int       esl_heap_IExtractTop(ESL_HEAP *hp, int *ret_val);

    int       esl_heap_IGetTop(ESL_HEAP *hp);



from libc.stdio cimport FILE

from libeasel cimport esl_pos_t, eslERRBUFSIZE


cdef extern from "esl_buffer.h" nogil:

    const size_t eslBUFFER_PAGESIZE
    const size_t eslBUFFER_SLURPSIZE

    cdef enum esl_buffer_mode_e:
        eslBUFFER_UNSET   = 0
        eslBUFFER_STREAM  = 1,
        eslBUFFER_CMDPIPE = 2,
        eslBUFFER_FILE    = 3,
        eslBUFFER_ALLFILE = 4,
        eslBUFFER_MMAP    = 5,
        eslBUFFER_STRING  = 6

    ctypedef struct ESL_BUFFER:
        char* mem
        esl_pos_t n
        esl_pos_t balloc
        esl_pos_t pos
        esl_pos_t baseoffset

        esl_pos_t anchor
        int nanchor

        FILE* fp
        char* filename
        char* cmdline

        esl_pos_t pagesize

        char errmsg[eslERRBUFSIZE]
        esl_buffer_mode_e mode_is

    # The ESL_BUFFER object: opening/closing
    int esl_buffer_Open      (const char *filename, const char *envvar, ESL_BUFFER **ret_bf);
    int esl_buffer_OpenFile  (const char *filename,                     ESL_BUFFER **ret_bf);
    int esl_buffer_OpenPipe  (const char *filename, const char *cmdfmt, ESL_BUFFER **ret_bf);
    int esl_buffer_OpenMem   (const char *p,         esl_pos_t  n,      ESL_BUFFER **ret_bf);
    int esl_buffer_OpenStream(FILE *fp,                                 ESL_BUFFER **ret_bf);
    int esl_buffer_Close(ESL_BUFFER *bf);

    # Positioning and anchoring an ESL_BUFFER
    esl_pos_t esl_buffer_GetOffset      (ESL_BUFFER *bf);
    int       esl_buffer_SetOffset      (ESL_BUFFER *bf, esl_pos_t offset);
    int       esl_buffer_SetAnchor      (ESL_BUFFER *bf, esl_pos_t offset);
    int       esl_buffer_SetStableAnchor(ESL_BUFFER *bf, esl_pos_t offset);
    int       esl_buffer_RaiseAnchor    (ESL_BUFFER *bf, esl_pos_t offset);

    # Raw access to the buffer
    int esl_buffer_Get(ESL_BUFFER *bf, char **ret_p, esl_pos_t *ret_n);
    int esl_buffer_Set(ESL_BUFFER *bf, char  *p,     esl_pos_t  nused);

    # Line-based parsing
    int esl_buffer_GetLine       (ESL_BUFFER *bf, char **opt_p, esl_pos_t *opt_n);
    int esl_buffer_FetchLine     (ESL_BUFFER *bf, char **opt_p, esl_pos_t *opt_n);
    int esl_buffer_FetchLineAsStr(ESL_BUFFER *bf, char **opt_s, esl_pos_t *opt_n);

    # Token-based parsing
    int esl_buffer_GetToken       (ESL_BUFFER *bf, const char *sep, char **opt_p, esl_pos_t *opt_n);
    int esl_buffer_FetchToken     (ESL_BUFFER *bf, const char *sep, char **opt_p, esl_pos_t *opt_n);
    int esl_buffer_FetchTokenAsStr(ESL_BUFFER *bf, const char *sep, char **opt_p, esl_pos_t *opt_n);

    # Binary (fread-like) parsing
    int esl_buffer_Read(ESL_BUFFER *bf, size_t nbytes, void *p);



from libc.stdint cimport int64_t
from libc.stdio cimport FILE
from posix.types cimport off_t

from libeasel cimport ESL_DSQ
from libeasel.alphabet cimport ESL_ALPHABET
from libeasel.sqio cimport ESL_SQFILE
from libeasel.sq cimport ESL_SQ, ESL_SQ_BLOCK


cdef extern from "reexports/esl_sqio_ascii.h":

    int   sqascii_GuessFileFormat(ESL_SQFILE *sqfp, int *ret_fmt)
    int   sqascii_Position       (ESL_SQFILE *sqfp, off_t offset)
    void  sqascii_Close          (ESL_SQFILE *sqfp)
    int   sqascii_SetDigital     (ESL_SQFILE *sqfp, const ESL_ALPHABET *abc)
    int   sqascii_GuessAlphabet  (ESL_SQFILE *sqfp, int *ret_type)
    int   sqascii_Read           (ESL_SQFILE *sqfp, ESL_SQ *sq)
    int   sqascii_ReadInfo       (ESL_SQFILE *sqfp, ESL_SQ *sq)
    int   sqascii_ReadSequence   (ESL_SQFILE *sqfp, ESL_SQ *sq)
    int   sqascii_ReadWindow     (ESL_SQFILE *sqfp, int C, int W, ESL_SQ *sq)
    int   sqascii_ReadBlock      (ESL_SQFILE *sqfp, ESL_SQ_BLOCK *sqBlock, int max_residues, int max_sequences, int max_init_window, int long_target)
    int   sqascii_Echo           (ESL_SQFILE *sqfp, const ESL_SQ *sq, FILE *ofp)

    int   sqascii_IsRewindable   (const ESL_SQFILE *sqfp)
    const char *sqascii_GetError (const ESL_SQFILE *sqfp)

    int   sqascii_OpenSSI         (ESL_SQFILE *sqfp, const char *ssifile_hint)
    int   sqascii_PositionByKey   (ESL_SQFILE *sqfp, const char *key)
    int   sqascii_PositionByNumber(ESL_SQFILE *sqfp, int which)
    int   sqascii_Fetch           (ESL_SQFILE *sqfp, const char *key, ESL_SQ *sq)
    int   sqascii_FetchInfo       (ESL_SQFILE *sqfp, const char *key, ESL_SQ *sq)
    int   sqascii_FetchSubseq     (ESL_SQFILE *sqfp, const char *source, int64_t start, int64_t end, ESL_SQ *sq)

    # Internal routines shared by parsers.
    int  loadmem  (ESL_SQFILE *sqfp) except -1
    int  loadbuf  (ESL_SQFILE *sqfp) except -1
    int  nextchar (ESL_SQFILE *sqfp, char *ret_c) except -1
    int  seebuf   (ESL_SQFILE *sqfp, int64_t maxn, int64_t *opt_nres, int64_t *opt_endpos) except -1
    void addbuf   (ESL_SQFILE *sqfp, ESL_SQ *sq, int64_t nres) except *
    void skipbuf  (ESL_SQFILE *sqfp, int64_t nskip) except *
    int  read_nres(ESL_SQFILE *sqfp, ESL_SQ *sq, int64_t nskip, int64_t nres, int64_t *opt_actual_nres) except -1
    int  skip_whitespace(ESL_SQFILE *sqfp) except -1

    # EMBL format also UniProt, TrEMBL
    void config_embl(ESL_SQFILE *sqfp) nogil
    void inmap_embl (ESL_SQFILE *sqfp, const ESL_DSQ *abc_inmap) nogil
    int  header_embl(ESL_SQFILE *sqfp, ESL_SQ *sq) except -1
    int  skip_embl  (ESL_SQFILE *sqfp, ESL_SQ *sq) except -1
    int  end_embl   (ESL_SQFILE *sqfp, ESL_SQ *sq) except -1

    # GenBank format also DDBJ
    void config_genbank(ESL_SQFILE *sqfp) nogil
    void inmap_genbank (ESL_SQFILE *sqfp, const ESL_DSQ *abc_inmap) nogil
    int  header_genbank(ESL_SQFILE *sqfp, ESL_SQ *sq) except -1
    int  skip_genbank  (ESL_SQFILE *sqfp, ESL_SQ *sq) except -1
    int  end_genbank   (ESL_SQFILE *sqfp, ESL_SQ *sq) except -1

    # FASTA format
    void config_fasta(ESL_SQFILE *sqfp) nogil
    void inmap_fasta (ESL_SQFILE *sqfp, const ESL_DSQ *abc_inmap) nogil
    int  header_fasta(ESL_SQFILE *sqfp, ESL_SQ *sq) except -1
    int  skip_fasta  (ESL_SQFILE *sqfp, ESL_SQ *sq) except -1
    int  end_fasta   (ESL_SQFILE *sqfp, ESL_SQ *sq) except -1

    # daemon format
    void config_daemon(ESL_SQFILE *sqfp) nogil
    void inmap_daemon (ESL_SQFILE *sqfp, const ESL_DSQ *abc_inmap) nogil
    int  end_daemon   (ESL_SQFILE *sqfp, ESL_SQ *sq) except -1

    # HMMPGMD format
    int  fileheader_hmmpgmd(ESL_SQFILE *sqfp) except -1



from libc.stdint cimport int64_t, uint32_t, uint64_t
from libc.stdio cimport FILE
from posix.types cimport off_t

from libeasel cimport ESL_DSQ
from libeasel.alphabet cimport ESL_ALPHABET
from libeasel.sq cimport ESL_SQ, ESL_SQ_BLOCK
from libeasel.sqio.ascii cimport ESL_SQASCII_DATA
from libeasel.sqio.ncbi cimport ESL_SQNCBI_DATA


cdef extern from "esl_sqio.h" nogil:

    ctypedef union ESL_SQDATA:
        ESL_SQASCII_DATA ascii
        ESL_SQNCBI_DATA  ncbi


    ctypedef esl_sqio_s ESL_SQFILE
    cdef struct esl_sqio_s:
        char *filename
        bint   do_digital
        const ESL_ALPHABET *abc

        int     format
        ESL_DSQ[128] inmap

        int   (*position)        (esl_sqio_s *sqfp, off_t offset)
        void  (*close)           (esl_sqio_s *sqfp)

        int   (*set_digital)     (esl_sqio_s *sqfp, const ESL_ALPHABET *abc)
        int   (*guess_alphabet)  (esl_sqio_s *sqfp, int *ret_type)

        int   (*read)            (esl_sqio_s *sqfp, ESL_SQ *sq)
        int   (*read_info)       (esl_sqio_s *sqfp, ESL_SQ *sq)
        int   (*read_seq)        (esl_sqio_s *sqfp, ESL_SQ *sq)
        int   (*read_window)     (esl_sqio_s *sqfp, int C, int W, ESL_SQ *sq)
        int   (*echo)            (esl_sqio_s *sqfp, const ESL_SQ *sq, FILE *ofp)

        int   (*read_block)      (esl_sqio_s *sqfp, ESL_SQ_BLOCK *sqBlock, int max_residues, int max_sequences, int max_init_window, int long_target)

        int   (*open_ssi)        (esl_sqio_s *sqfp, const char *ssifile_hint)
        int   (*pos_by_key)      (esl_sqio_s *sqfp, const char *key)
        int   (*pos_by_number)   (esl_sqio_s *sqfp, int which)

        int   (*fetch)           (esl_sqio_s *sqfp, const char *key, ESL_SQ *sq)
        int   (*fetch_info)      (esl_sqio_s *sqfp, const char *key, ESL_SQ *sq)
        int   (*fetch_subseq)    (esl_sqio_s *sqfp, const char *source, int64_t start, int64_t end, ESL_SQ *sq)

        int   (*is_rewindable)   (const esl_sqio_s *sqfp)
        const char *(*get_error) (const esl_sqio_s *sqfp)

        ESL_SQDATA data


    ctypedef esl_sqcache_s ESL_SQCACHE
    cdef struct esl_sqcache_s:
        char* filename
        int format
        const ESL_ALPHABET* abc
        uint32_t seq_count
        uint64_t res_count
        uint32_t max_seq
        ESL_SQ* sq_list
        void* residue_mem
        void* header_mem
        uint64_t res_size
        uint64_t hdr_size


    cdef enum:
        eslSQFILE_UNKNOWN = 0
        eslSQFILE_FASTA = 1
        eslSQFILE_EMBL = 2
        eslSQFILE_GENBANK = 3
        eslSQFILE_DDBJ = 4
        eslSQFILE_UNIPROT = 5
        eslSQFILE_NCBI = 6
        eslSQFILE_DAEMON = 7
        eslSQFILE_HMMPGMD = 8
        eslSQFILE_FMINDEX = 9


    const size_t eslREADBUFSIZE


    int  esl_sqfile_Open(const char *seqfile, int fmt, const char *env, ESL_SQFILE **ret_sqfp)# except *
    void esl_sqfile_Close(ESL_SQFILE *sqfp)

    int  esl_sqfile_OpenDigital(const ESL_ALPHABET *abc, const char *filename, int format, const char *env, ESL_SQFILE **ret_sqfp)# except *
    int  esl_sqfile_SetDigital(ESL_SQFILE *sqfp, const ESL_ALPHABET *abc)# except *
    int  esl_sqfile_GuessAlphabet(ESL_SQFILE *sqfp, int *ret_type)# except *

    int   esl_sqio_Read        (ESL_SQFILE *sqfp, ESL_SQ *sq) #except *
    int   esl_sqio_ReadInfo    (ESL_SQFILE *sqfp, ESL_SQ *sq) #except *
    int   esl_sqio_ReadWindow  (ESL_SQFILE *sqfp, int C, int W, ESL_SQ *sq) #except *
    int   esl_sqio_ReadSequence(ESL_SQFILE *sqfp, ESL_SQ *sq) #except *
    int   esl_sqio_ReadBlock   (ESL_SQFILE *sqfp, ESL_SQ_BLOCK *sqBlock, int max_residues, int max_sequences, int max_init_window, int long_target) #except *
    int   esl_sqio_Parse       (char *buffer, int size, ESL_SQ *s, int format) #except *

    int   esl_sqio_Write       (FILE *fp, ESL_SQ *s, int format, int update)# except *
    int   esl_sqio_Echo        (ESL_SQFILE *sqfp, const ESL_SQ *sq, FILE *ofp)# except *

    const char  *esl_sqfile_GetErrorBuf(const ESL_SQFILE *sqfp)
    int   esl_sqfile_IsRewindable(const ESL_SQFILE *sqfp)
    int   esl_sqio_IsAlignment(int fmt)
    int   esl_sqio_EncodeFormat(char *fmtstring)
    char *esl_sqio_DecodeFormat(int fmt)
    int   esl_sqfile_Position(ESL_SQFILE *sqfp, off_t offset)# except *
    int   esl_sqio_Ignore(ESL_SQFILE *sqfp, const char *ignoredchars)
    int   esl_sqio_AcceptAs(ESL_SQFILE *sqfp, char *xchars, char readas)

    int   esl_sqfile_OpenSSI         (ESL_SQFILE *sqfp, const char *ssifile_hint)# except *
    int   esl_sqfile_PositionByKey   (ESL_SQFILE *sqfp, const char *key)# except *
    int   esl_sqfile_PositionByNumber(ESL_SQFILE *sqfp, int which)# except *

    int   esl_sqio_Fetch      (ESL_SQFILE *sqfp, const char *key, ESL_SQ *sq)# except *
    int   esl_sqio_FetchInfo  (ESL_SQFILE *sqfp, const char *key, ESL_SQ *sq)# except *
    int   esl_sqio_FetchSubseq(ESL_SQFILE *sqfp, const char *source, int64_t start, int64_t end, ESL_SQ *sq)# except *

    int   esl_sqfile_Cache(const ESL_ALPHABET *abc, const char *seqfile, int fmt, const char *env, ESL_SQCACHE **ret_sqcache)# except *
    void  esl_sqfile_Free(ESL_SQCACHE *sqcache)



from libc.stdio cimport FILE
from posix.types cimport off_t

from libeasel cimport eslERRBUFSIZE
from libeasel.alphabet cimport ESL_ALPHABET
from libeasel.fileparser cimport ESL_FILEPARSER
from libeasel.ssi cimport ESL_SSI
from libhmmer.p7_hmm cimport P7_HMM


cdef extern from "hmmer.h" nogil:


    cdef enum p7_hmmfile_formats_e:
        p7_HMMFILE_20 = 0
        p7_HMMFILE_3a = 1
        p7_HMMFILE_3b = 2
        p7_HMMFILE_3c = 3
        p7_HMMFILE_3d = 4
        p7_HMMFILE_3e = 5
        p7_HMMFILE_3f = 6


    ctypedef p7_hmmfile_s P7_HMMFILE
    cdef struct p7_hmmfile_s:
        FILE* f
        char* fname
        ESL_SSI* ssi

        bint do_gzip
        bint do_stdin
        bint newly_opened
        bint is_pressed

        int format
        int (*parser)(p7_hmmfile_s*, ESL_ALPHABET**, P7_HMM**)
        ESL_FILEPARSER* efp

        FILE* ffp
        FILE* pfp

        # int syncRead
        # pthread_mutex_t readMutex

        char[eslERRBUFSIZE] errbuf


    int  p7_hmmfile_Open     (const char *filename, char *env, P7_HMMFILE **ret_hfp, char *errbuf)# except *
    int  p7_hmmfile_OpenNoDB (const char *filename, char *env, P7_HMMFILE **ret_hfp, char *errbuf)# except *
    int  p7_hmmfile_OpenBuffer(const char *buffer, int size, P7_HMMFILE **ret_hfp)# except *
    void p7_hmmfile_Close(P7_HMMFILE *hfp)# except *

    int  p7_hmmfile_WriteBinary(FILE *fp, int format, P7_HMM *hmm)# except *
    int  p7_hmmfile_WriteASCII (FILE *fp, int format, P7_HMM *hmm)# except *
    int  p7_hmmfile_WriteToString (char **s, int format, P7_HMM *hmm)

    int  p7_hmmfile_Read(P7_HMMFILE *hfp, ESL_ALPHABET **ret_abc,  P7_HMM **opt_hmm)# except *
    int  p7_hmmfile_PositionByKey(P7_HMMFILE *hfp, const char *key)
    int  p7_hmmfile_Position(P7_HMMFILE *hfp, const off_t offset)



from libc.stdint cimport uint8_t, uint32_t, uint64_t, int64_t

from libeasel.alphabet cimport ESL_ALPHABET
from libeasel.getopts import ESL_GETOPTS
from libeasel.sq cimport ESL_SQ
from libeasel.rand64 cimport ESL_RAND64

from libhmmer.p7_pipeline cimport p7_zsetby_e
from libhmmer.p7_hmm cimport P7_HMM


cdef extern from "hmmpgmd.h" nogil:

    ctypedef struct HMMD_SEARCH_STATUS:
        uint32_t status
        uint64_t msg_size

    ctypedef struct HMMD_SEARCH_STATS:
        double   elapsed
        double   user
        double   sys

        double   Z
        double   domZ

        p7_zsetby_e Z_setby
        p7_zsetby_e domZ_setby

        uint64_t    nmodels
        uint64_t    nseqs
        uint64_t    n_past_msv
        uint64_t    n_past_bias
        uint64_t    n_past_vit
        uint64_t    n_past_fwd

        uint64_t    nhits
        uint64_t    nreported
        uint64_t    nincluded
        uint64_t*   hit_offsets


    const uint32_t HMMD_SEQUENCE
    const uint32_t HMMD_HMM

    const uint32_t HMMD_CMD_SEARCH
    const uint32_t HMMD_CMD_SCAN
    const uint32_t HMMD_CMD_INIT
    const uint32_t HMMD_CMD_SHUTDOWN

    const uint32_t MAX_INIT_DESC


    ctypedef struct HMMD_SEARCH_CMD:
        uint32_t  db_indx
        uint32_t  db_type
        uint32_t  inx
        uint32_t  cnt
        uint32_t  query_type
        uint32_t  query_length
        uint32_t  opts_length
        char      data[1]

    ctypedef struct HMMD_INIT_CMD:
        char      sid[MAX_INIT_DESC]
        char      hid[MAX_INIT_DESC]
        uint32_t  seqdb_off
        uint32_t  hmmdb_off
        uint32_t  db_cnt
        uint32_t  seq_cnt
        uint32_t  hmm_cnt
        uint32_t  model_cnt
        char      data[1]

    ctypedef struct HMMD_INIT_RESET:
        char      ip_addr[1]

    ctypedef struct HMMD_HEADER:
        uint32_t  length
        uint32_t  command
        uint32_t  status

    ctypedef struct HMMD_COMMAND:
        HMMD_HEADER   hdr
        HMMD_INIT_CMD init
        HMMD_SEARCH_CMD srch
        HMMD_INIT_RESET reset

    const size_t HMMD_SEARCH_STATUS_SERIAL_SIZE
    const size_t HMMD_SEARCH_STATS_SERIAL_BASE

    cdef struct queue_data_s:
        uint32_t  cmd_type
        uint32_t  query_type
        P7_HMM*   hmm
        ESL_SQ*   seq
        ESL_ALPHABET* abc
        # ESL_GETOPTS*  opts
        HMMD_COMMAND* cmd

        int  sock
        char ip_addr[64]

        int  dbx
        int  inx
        int  cnt
    ctypedef queue_data_s QUEUE_DATA

    ctypedef struct RANGE_LIST:
        int N
        uint32_t* starts
        uint32_t* ends

    cdef void free_QueueData(QUEUE_DATA* data)
    cdef int  hmmpgmd_IsWithinRanges(int64_t sq_idx, RANGE_LIST* list)
    cdef int  hmmpgmd_GetRanges(RANGE_LIST* list, char* rangestr)

    # cdef int process_searchopts(int fd, char* cmdstr, ESL_GETOPTS** ret_opts)

    # cdef void worker_process(ESL_GETOPTS* go)
    # cdef void master_process(ESL_GETOPTS* go)

    cdef int p7_hmmd_search_stats_Serialize(const HMMD_SEARCH_STATS *obj, uint8_t **buf, uint32_t *n, uint32_t *nalloc)
    cdef int p7_hmmd_search_stats_Deserialize(const uint8_t *buf, uint32_t *pos, HMMD_SEARCH_STATS *ret_obj)

    cdef int hmmd_search_status_Serialize(const HMMD_SEARCH_STATUS *obj, uint8_t **buf, uint32_t *n, uint32_t *nalloc)
    cdef int hmmd_search_status_Deserialize(const uint8_t *buf, uint32_t *n, HMMD_SEARCH_STATUS *ret_obj)
    cdef int hmmd_search_status_TestSample(ESL_RAND64 *rng, HMMD_SEARCH_STATUS **ret_obj)
    cdef int hmmd_search_status_Compare(HMMD_SEARCH_STATUS *first, HMMD_SEARCH_STATUS *second)



import copy
import datetime
import io
import itertools
import os
import shutil
import sys
import unittest
import tempfile
from unittest import mock

import pyhmmer
from pyhmmer.errors import EaselError, AlphabetMismatch
from pyhmmer.easel import Alphabet, SequenceFile, MSAFile
from pyhmmer.plan7 import HMMFile, Pipeline, Builder, Background

from .. import __name__ as __package__
from .utils import HMMER_FOLDER, resource_files


@unittest.skipUnless(resource_files, "importlib.resources.files not available")
class _TestBuilderBase(object):

    @classmethod
    def setUpClass(cls):
        cls.testdata = os.path.join(HMMER_FOLDER, "testsuite")

        cls.faa_path = resource_files(__package__).joinpath("data", "seqs", "PKSI.faa")
        with SequenceFile(cls.faa_path, digital=True) as seqf:
            cls.proteins = list(seqf)

        if os.path.exists(cls.testdata):
            cls.ecori_fa = os.path.join(cls.testdata, "ecori.fa")
            with SequenceFile(cls.ecori_fa) as seqf:
                cls.dna = next(seqf)

            cls.ecori_hmm = cls.ecori_fa.replace(".fa", ".hmm")
            with HMMFile(cls.ecori_hmm) as hmmf:
                cls.ecori = next(hmmf)

            cls.msa_path = os.path.join(cls.testdata, "3box.sto")
            with MSAFile(cls.msa_path, digital=True) as msa_file:
                cls.msa = next(msa_file)
                cls.msa.name = b"3box"


class TestBuilder(unittest.TestCase):

    def test_invalid_window_length(self):
        abc = Alphabet.amino()
        builder = Builder(alphabet=abc)
        with self.assertRaises(ValueError):
            builder.window_length = 2

    def test_init_error(self):
        abc = Alphabet.amino()
        self.assertRaises(ValueError, Builder, alphabet=abc, architecture="nonsense")
        self.assertRaises(ValueError, Builder, alphabet=abc, weighting="nonsense")
        self.assertRaises(ValueError, Builder, alphabet=abc, effective_number="nonsense")
        self.assertRaises(ValueError, Builder, alphabet=abc, prior_scheme="nonsense")

    def test_init_effective_number(self):
        abc = Alphabet.amino()
        builder = Builder(alphabet=abc, effective_number=2.0)
        self.assertEqual(builder.effective_number, 2.0)
        self.assertRaises(TypeError, Builder, alphabet=abc, effective_number=[])

    def test_copy(self):
        abc = Alphabet.amino()
        builder = Builder(alphabet=abc)
        bc1 = builder.copy()
        self.assertEqual(builder.alphabet, bc1.alphabet)
        self.assertEqual(builder.seed, bc1.seed)
        bc2 = copy.copy(bc1)
        self.assertEqual(builder.alphabet, bc2.alphabet)
        self.assertEqual(builder.seed, bc2.seed)


class TestBuilderSingle(_TestBuilderBase, unittest.TestCase):

    def test_alphabet_mismatch(self):
        abc = Alphabet.dna()
        amino = Alphabet.amino()

        builder = Builder(alphabet=abc)
        bg = Background(alphabet=abc)

        seq = self.proteins[0]
        self.assertRaises(AlphabetMismatch, builder.build, seq, bg)

    def test_command_line(self):
        abc = Alphabet.amino()
        builder = Builder(alphabet=abc)
        seq = self.proteins[1]

        argv = ["mycommand", "--param", "1"]
        with mock.patch.object(sys, "argv", argv):
            hmm, profile, opt = builder.build(seq, Background(abc))
        self.assertEqual(hmm.command_line, " ".join(argv))

    def test_creation_time(self):
        abc = Alphabet.amino()
        builder = Builder(alphabet=abc)
        seq = self.proteins[1]

        hmm, profile, opt = builder.build(seq, Background(abc))
        self.assertIsInstance(hmm.creation_time, datetime.datetime)

    def test_protein(self):
        abc = Alphabet.amino()
        builder = Builder(alphabet=abc)
        seq = self.proteins[1]
        hmm, profile, opt = builder.build(seq, Background(abc))
        self.assertEqual(hmm.name, seq.name)
        self.assertEqual(hmm.alphabet, abc)
        self.assertEqual(profile.alphabet, abc)
        self.assertEqual(opt.alphabet, abc)

    @unittest.skipUnless(os.path.exists(HMMER_FOLDER), "test data not available")
    def test_dna(self):
        abc = Alphabet.dna()
        builder = Builder(alphabet=abc)
        hmm, profile, opt = builder.build(self.dna.digitize(abc), Background(abc))
        self.assertEqual(hmm.name, self.dna.name)
        self.assertEqual(hmm.alphabet, abc)
        self.assertEqual(profile.alphabet, abc)
        self.assertEqual(opt.alphabet, abc)
        self.assertEqual(hmm.M, self.ecori.M)

    @unittest.skipUnless(os.path.exists(HMMER_FOLDER), "test data not available")
    def test_score_system_change(self):
        abc = Alphabet.amino()
        bg = Background(abc)
        builder = Builder(alphabet=abc)
        seq = self.proteins[1]
        # build first HMM with mx=BLOSUM62, popen=0.02, pextend=0.4
        self.assertEqual(builder.score_matrix, "BLOSUM62")
        hmm1, _, _ = builder.build(seq, bg)
        hmm1.command_line = hmm1.creation_time = None
        # build second HMM with mx=BLOSUM90, popen=0.02, pextend=0.4
        # --> the transition probabilities won't change, match emissions will
        builder.score_matrix = "BLOSUM90"
        self.assertEqual(builder.score_matrix, "BLOSUM90")
        hmm2, _, _ = builder.build(seq, bg)
        hmm2.command_line = hmm1.creation_time = None
        self.assertEqual(hmm1.transition_probabilities, hmm2.transition_probabilities)
        self.assertNotEqual(hmm1.match_emissions, hmm2.match_emissions)
        # build third HMM with mx=BLOSUM90, popen=0.04, pextend=0.2
        # --> the transition probabilities will change, match emissions won't
        builder.popen = 0.04
        builder.pextend = 0.2
        hmm3, _, _ = builder.build(seq, bg)
        hmm3.command_line = hmm1.creation_time = None
        self.assertNotEqual(hmm2.transition_probabilities, hmm3.transition_probabilities)
        self.assertEqual(hmm2.match_emissions, hmm3.match_emissions)


class TestBuilderMSA(_TestBuilderBase, unittest.TestCase):

    @unittest.skipUnless(os.path.exists(HMMER_FOLDER), "test data not available")
    def test_command_line(self):
        abc = Alphabet.dna()
        builder = Builder(alphabet=abc)

        with MSAFile(os.path.join(self.testdata, "3box.sto"), digital=True) as msa_file:
            msa = next(msa_file)
            msa.name = b"3box"

        with HMMFile(os.path.join(self.testdata, "3box.hmm")) as hmm_file:
            hmm_exp = next(hmm_file)

        argv = ["mycommand", "--param", "1"]
        with mock.patch.object(sys, "argv", argv):
            hmm, profile, opt = builder.build_msa(msa, Background(abc))
        self.assertEqual(hmm.command_line, " ".join(argv))

    @unittest.skipUnless(os.path.exists(HMMER_FOLDER), "test data not available")
    def test_creation_time(self):
        abc = Alphabet.dna()
        builder = Builder(alphabet=abc)

        hmm, profile, opt = builder.build_msa(self.msa, Background(abc))

        with HMMFile(os.path.join(self.testdata, "3box.hmm")) as hmm_file:
            hmm_exp = next(hmm_file)

        self.assertIsInstance(hmm.creation_time, datetime.datetime)

    @unittest.skipUnless(os.path.exists(HMMER_FOLDER), "test data not available")
    def test_dna(self):
        abc = Alphabet.dna()
        builder = Builder(alphabet=abc)

        hmm, profile, opt = builder.build_msa(self.msa, Background(abc))

        with HMMFile(os.path.join(self.testdata, "3box.hmm")) as hmm_file:
            hmm_exp = next(hmm_file)

        self.assertEqual(hmm.name, b"3box")
        self.assertEqual(hmm.M, hmm_exp.M)

    @unittest.skipUnless(os.path.exists(HMMER_FOLDER), "test data not available")
    def test_alphabet_mismatch(self):
        dna = Alphabet.dna()
        amino = Alphabet.amino()

        builder = Builder(alphabet=dna)
        bg = Background(alphabet=amino)
        self.assertRaises(AlphabetMismatch, builder.build_msa, self.msa, bg)

        builder = Builder(alphabet=amino)
        bg = Background(alphabet=amino)
        self.assertRaises(AlphabetMismatch, builder.build_msa, self.msa, bg)

    def test_luxc(self):
        # create a new protein builder
        abc = Alphabet.amino()
        builder = Builder(alphabet=abc)
        # open the MSA
        msa_path = resource_files(__package__).joinpath("data", "msa", "LuxC.sto")
        with MSAFile(msa_path, digital=True, alphabet=abc) as msa_file:
            msa = msa_file.read()
            msa.name = b"LuxC"
        # read the expected HMM
        hmm_path = resource_files(__package__).joinpath("data", "hmms", "txt", "LuxC.hmm")
        with HMMFile(hmm_path) as hmm_file:
            hmm_hmmbuild = hmm_file.read()
            hmm_hmmbuild.creation_time = None
        # build the HMM from the MSA
        hmm_pyhmmer, _, _ = builder.build_msa(msa, Background(abc))
        hmm_pyhmmer.command_line = hmm_pyhmmer.creation_time = None
        # check M and the transition probabilities (which were the issue in v0.4.9)
        # NOTE: exact equality cannot be checked because the `hmmbuild` HMM was
        #       serialized, which drops a bit of numerical accuracy.
        self.assertEqual(hmm_pyhmmer.M, hmm_hmmbuild.M)
        for i in range(hmm_pyhmmer.M):
            for t1, t2 in zip(hmm_pyhmmer.transition_probabilities[i], hmm_hmmbuild.transition_probabilities[i]):
                self.assertAlmostEqual(t1, t2, places=5)
            for t1, t2 in zip(hmm_pyhmmer.insert_emissions[i], hmm_hmmbuild.insert_emissions[i]):
                self.assertAlmostEqual(t1, t2, places=5)
            for t1, t2 in zip(hmm_pyhmmer.match_emissions[i], hmm_hmmbuild.match_emissions[i]):
                self.assertAlmostEqual(t1, t2, places=5)


import copy
import io
import itertools
import os
import shutil
import unittest
import tempfile

import pyhmmer
from pyhmmer.errors import EaselError, AlphabetMismatch
from pyhmmer.easel import Alphabet, SequenceFile, DigitalMSA, TextMSA, DigitalSequenceBlock
from pyhmmer.plan7 import HMM, HMMFile, TraceAligner, Traces

from .. import __name__ as __package__
from .utils import HMMER_FOLDER, resource_files


@unittest.skipUnless(resource_files, "importlib.resources.files not available")
class TestTraceAligner(unittest.TestCase):

    def setUp(self):
        hmm_path = resource_files(__package__).joinpath("data", "hmms", "txt", "LuxC.hmm")
        seqs_path = resource_files(__package__).joinpath("data", "seqs", "LuxC.faa")

        with HMMFile(hmm_path) as hmm_file:
            self.hmm = next(hmm_file)
        with SequenceFile(seqs_path, digital=True, alphabet=self.hmm.alphabet) as seqs_file:
            self.seqs = seqs_file.read_block()

    def test_align_traces(self):
        # hmmalign tests/data/hmms/txt/LuxC.hmm tests/data/seqs/LuxC.faa
        aligner = TraceAligner()
        traces = aligner.compute_traces(self.hmm, self.seqs)
        self.assertEqual(len(traces), len(self.seqs))
        msa = aligner.align_traces(self.hmm, self.seqs, traces, all_consensus_cols=True)
        self.assertEqual(len(msa.sequences), len(self.seqs))
        self.assertEqual(len(msa), 567)

    def test_align_traces_trim(self):
        # hmmalign --trim tests/data/hmms/txt/LuxC.hmm tests/data/seqs/LuxC.faa
        aligner = TraceAligner()
        traces = aligner.compute_traces(self.hmm, self.seqs)
        self.assertEqual(len(traces), len(self.seqs))
        msa = aligner.align_traces(self.hmm, self.seqs, traces, all_consensus_cols=True, trim=True)
        self.assertEqual(len(msa.sequences), len(self.seqs))
        self.assertEqual(len(msa), 429)

    def test_align_traces_mismatch(self):
        aligner = TraceAligner()
        traces = Traces()
        self.assertRaises(ValueError, aligner.align_traces, self.hmm, self.seqs, traces)

    def test_align_traces_msa_type(self):
        aligner = TraceAligner()
        traces = Traces()
        seqs = DigitalSequenceBlock(self.hmm.alphabet)
        msa = aligner.align_traces(self.hmm, seqs, traces)
        self.assertIsInstance(msa, TextMSA)
        msa_d = aligner.align_traces(self.hmm, seqs, traces, digitize=True)
        self.assertIsInstance(msa_d, DigitalMSA)



from libc.stdio cimport FILE

from libeasel.random cimport ESL_RANDOMNESS
from libeasel.sq cimport ESL_SQ
from libhmmer.p7_domain cimport P7_DOMAIN
from libhmmer.p7_spensemble cimport P7_SPENSEMBLE
from libhmmer.p7_trace cimport P7_TRACE
from libhmmer.p7_alidisplay cimport P7_ALIDISPLAY

if HMMER_IMPL == "VMX":
    from libhmmer.impl_vmx.p7_omx cimport P7_OMX
    from libhmmer.impl_vmx.p7_oprofile cimport P7_OPROFILE
elif HMMER_IMPL == "SSE":
    from libhmmer.impl_sse.p7_omx cimport P7_OMX
    from libhmmer.impl_sse.p7_oprofile cimport P7_OPROFILE
elif HMMER_IMPL == "NEON":
    from libhmmer.impl_neon.p7_omx cimport P7_OMX
    from libhmmer.impl_neon.p7_oprofile cimport P7_OPROFILE


cdef extern from "hmmer.h" nogil:

    cdef struct p7_domaindef_s:

        float* mocc
        float* btot
        float* etot
        int L
        int Lalloc

        float* n2sc

        ESL_RANDOMNESS* r
        int do_reseeding
        P7_SPENSEMBLE* sp
        P7_TRACE* tr
        P7_TRACE *gtr

        float rt1
        float rt2
        float rt3

        int nsamples
        float min_overlap
        int of_smaller
        int max_diagdiff
        float min_posterior
        float min_endpointp

        P7_DOMAIN* dcl
        int ndom
        int nalloc

        float nexpected
        int nregions
        int nclustered
        int noverlaps
        int nenvelopes
    ctypedef p7_domaindef_s P7_DOMAINDEF


    P7_DOMAINDEF *p7_domaindef_Create (ESL_RANDOMNESS *r)
    int           p7_domaindef_Fetch  (P7_DOMAINDEF *ddef, int which, int *opt_i, int *opt_j, float *opt_sc, P7_ALIDISPLAY **opt_ad)
    int           p7_domaindef_GrowTo (P7_DOMAINDEF *ddef, int L)
    int           p7_domaindef_Reuse  (P7_DOMAINDEF *ddef)
    int           p7_domaindef_DumpPosteriors(FILE *ofp, P7_DOMAINDEF *ddef)
    void          p7_domaindef_Destroy(P7_DOMAINDEF *ddef)

    # int p7_domaindef_ByViterbi            (P7_PROFILE *gm, const ESL_SQ *sq, const ESL_SQ *ntsq, P7_GMX *gx1, P7_GMX *gx2, P7_DOMAINDEF *ddef)
    # int p7_domaindef_ByPosteriorHeuristics(const ESL_SQ *sq, const ESL_SQ *ntsq, P7_OPROFILE *om, P7_OMX *oxf, P7_OMX *oxb, P7_OMX *fwd, P7_OMX *bck,
    # 				                                  P7_DOMAINDEF *ddef, P7_BG *bg, int long_target,
    # 				                                  P7_BG *bg_tmp, float *scores_arr, float *fwd_emissions_arr);



from libc.stdint cimport uint64_t, int64_t
from libc.stdio cimport FILE


cdef extern from "esl_rand64.h" nogil:

    ctypedef struct ESL_RAND64:
        int mti
        uint64_t[312] mt
        uint64_t seed


    ESL_RAND64 *esl_rand64_Create (uint64_t seed)
    int         esl_rand64_Init   (ESL_RAND64 *rng, uint64_t seed)
    uint64_t    esl_rand64_GetSeed(ESL_RAND64 *rng)
    void        esl_rand64_Destroy(ESL_RAND64 *rng)

    uint64_t    esl_rand64(ESL_RAND64 *rng)
    uint64_t    esl_rand64_Roll(ESL_RAND64 *rng, uint64_t n);
    double      esl_rand64_double(ESL_RAND64 *rng);
    double      esl_rand64_double_closed(ESL_RAND64 *rng);
    double      esl_rand64_double_open(ESL_RAND64 *rng);

    int         esl_rand64_Deal(ESL_RAND64 *rng, int64_t m, int64_t n, int64_t *deal)

    int         esl_rand64_Dump(FILE *fp, ESL_RAND64 *rng)



# coding: utf-8
import collections.abc
import datetime
import enum
import os
import sys
import types
import typing
from typing import Any

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal  # type: ignore

from .utils import SizedIterator
from .easel import (
    Alphabet,
    Sequence,
    DigitalSequence,
    DigitalSequenceBlock,
    KeyHash,
    MSA,
    DigitalMSA,
    TextMSA,
    Randomness,
    SequenceFile,
    VectorF,
    VectorU8,
    MatrixF,
    MatrixU8,
)

BIT_CUTOFFS = Literal["gathering", "trusted", "noise"]
SORT_KEY = Literal["key", "seqidx"]
ARCHITECTURE = Literal["fast", "hand"]
WEIGHTING = Literal["pb", "gsc", "blosum", "none", "given"]
EFFECTIVE = Literal["entropy", "exp", "clust", "none"]
PRIOR_SCHEME = Literal["laplace", "alphabet"]
STRAND = Literal["watson", "crick"]
STRAND_SIGN = Literal["+", "-"]
HITS_FORMAT = Literal["targets", "domain", "pfam"]
HITS_MODE = Literal["search", "scan"]

Q = typing.TypeVar("Q")  # pipeline query type

class Alignment(collections.abc.Sized):
    domain: Domain
    def __len__(self) -> int: ...
    def __getstate__(self) -> typing.Dict[str, object]: ...
    def __setstate__(self, state: typing.Dict[str, object]) -> None: ...
    @property
    def hmm_accession(self) -> bytes: ...
    @property
    def hmm_from(self) -> int: ...
    @property
    def hmm_length(self) -> int: ...
    @property
    def hmm_name(self) -> bytes: ...
    @property
    def hmm_sequence(self) -> str: ...
    @property
    def hmm_to(self) -> int: ...
    @property
    def posterior_probabilities(self) -> str: ...
    @property
    def target_from(self) -> int: ...
    @property
    def target_length(self) -> int: ...
    @property
    def target_name(self) -> bytes: ...
    @property
    def target_sequence(self) -> str: ...
    @property
    def target_to(self) -> int: ...
    @property
    def identity_sequence(self) -> str: ...

class Background(object):
    def __init__(self, alphabet: Alphabet, uniform: bool = False) -> None: ...
    def __copy__(self) -> Background: ...
    @property
    def L(self) -> int: ...
    @L.setter
    def L(self, L: int) -> None: ...
    @property
    def residue_frequencies(self) -> VectorF: ...
    @property
    def transition_probability(self) -> float: ...
    @property
    def omega(self) -> float: ...
    @omega.setter
    def omega(self, omega: float) -> None: ...
    def copy(self) -> Background: ...

class Builder(object):
    _ARCHITECTURE_STRATEGY: typing.ClassVar[typing.Dict[str, int]]
    _WEIGHTING_STRATEGY: typing.ClassVar[typing.Dict[str, int]]
    _EFFECTIVE_STRATEGY: typing.ClassVar[typing.Dict[str, int]]
    alphabet: Alphabet
    randomness: Randomness
    score_matrix: str
    architecture: ARCHITECTURE
    weighting: WEIGHTING
    effective_number: typing.Union[EFFECTIVE, int, float]
    prior_scheme: typing.Optional[PRIOR_SCHEME]
    popen: float
    pextend: float
    def __init__(
        self,
        alphabet: Alphabet,
        *,
        architecture: ARCHITECTURE = "fast",
        weighting: WEIGHTING = "pb",
        effective_number: typing.Union[EFFECTIVE, int, float] = "entropy",
        prior_scheme: typing.Optional[PRIOR_SCHEME] = "alphabet",
        symfrac: float = 0.5,
        fragthresh: float = 0.5,
        wid: float = 0.62,
        esigma: float = 45.0,
        eid: float = 0.62,
        EmL: int = 200,
        EmN: int = 200,
        EvL: int = 200,
        EvN: int = 200,
        EfL: int = 100,
        EfN: int = 200,
        Eft: float = 0.04,
        seed: int = 42,
        ere: typing.Optional[float] = None,
        popen: typing.Optional[float] = None,
        pextend: typing.Optional[float] = None,
        score_matrix: typing.Optional[str] = None,
        window_length: typing.Optional[int] = None,
        window_beta: typing.Optional[float] = None,
    ) -> None: ...
    def __copy__(self) -> Builder: ...
    @property
    def seed(self) -> int: ...
    @seed.setter
    def seed(self, seed: int) -> None: ...
    @property
    def window_length(self) -> typing.Optional[int]: ...
    @window_length.setter
    def window_length(self, window_length: typing.Optional[int]) -> None: ...
    @property
    def window_beta(self) -> typing.Optional[float]: ...
    @window_beta.setter
    def window_beta(self, window_beta: typing.Optional[float]) -> None: ...
    def build(
        self,
        sequence: DigitalSequence,
        background: Background,
    ) -> typing.Tuple[HMM, Profile, OptimizedProfile]: ...
    def build_msa(
        self,
        msa: DigitalMSA,
        background: Background,
    ) -> typing.Tuple[HMM, Profile, OptimizedProfile]: ...
    def copy(self) -> Builder: ...

class Cutoffs(object):
    def __init__(self, owner: typing.Union[Profile, HMM]) -> None: ...
    def __copy__(self) -> Cutoffs: ...
    def __str__(self) -> str: ...
    def __eq__(self, other: object) -> bool: ...
    @property
    def gathering(self) -> typing.Optional[typing.Tuple[float, float]]: ...
    @gathering.setter
    def gathering(
        self, gathering: typing.Optional[typing.Tuple[float, float]]
    ) -> None: ...
    @gathering.deleter
    def gathering(self) -> None: ...
    @property
    def gathering1(self) -> typing.Optional[float]: ...
    @property
    def gathering2(self) -> typing.Optional[float]: ...
    @property
    def trusted(self) -> typing.Optional[typing.Tuple[float, float]]: ...
    @trusted.setter
    def trusted(self, trusted: typing.Optional[typing.Tuple[float, float]]) -> None: ...
    @trusted.deleter
    def trusted(self) -> None: ...
    @property
    def trusted1(self) -> typing.Optional[float]: ...
    @property
    def trusted2(self) -> typing.Optional[float]: ...
    @property
    def noise(self) -> typing.Optional[typing.Tuple[float, float]]: ...
    @noise.setter
    def noise(self, noise: typing.Optional[typing.Tuple[float, float]]) -> None: ...
    @noise.deleter
    def noise(self) -> None: ...
    @property
    def noise1(self) -> typing.Optional[float]: ...
    @property
    def noise2(self) -> typing.Optional[float]: ...
    def gathering_available(self) -> bool: ...
    def trusted_available(self) -> bool: ...
    def noise_available(self) -> bool: ...
    def as_vector(self) -> VectorF: ...

class Domain(object):
    alignment: Alignment
    hit: Hit
    def __getstate__(self) -> typing.Dict[str, object]: ...
    def __setstate__(self, state: typing.Dict[str, object]) -> None: ...
    @property
    def env_from(self) -> int: ...
    @property
    def env_to(self) -> int: ...
    @property
    def strand(self) -> typing.Optional[STRAND_SIGN]: ...
    @property
    def score(self) -> float: ...
    @property
    def bias(self) -> float: ...
    @property
    def correction(self) -> float: ...
    @property
    def envelope_score(self) -> float: ...
    @property
    def c_evalue(self) -> float: ...
    @property
    def i_evalue(self) -> float: ...
    @property
    def pvalue(self) -> float: ...
    @property
    def included(self) -> bool: ...
    @included.setter
    def included(self, included: bool) -> None: ...
    @property
    def reported(self) -> bool: ...
    @reported.setter
    def reported(self, reported: bool) -> None: ...

class Domains(typing.Sequence[Domain]):
    hit: Hit
    def __len__(self) -> int: ...
    def __getstate__(self) -> typing.List[typing.Dict[str, object]]: ...
    @typing.overload
    def __getitem__(self, index: int) -> Domain: ...
    @typing.overload
    def __getitem__(self, index: slice) -> typing.Sequence[Domain]: ...
    @property
    def reported(self) -> SizedIterator[Domain]: ...
    @property
    def included(self) -> SizedIterator[Domain]: ...

class EvalueParameters:
    def __init__(self, owner: typing.Union[Profile, HMM]) -> None: ...
    def __copy__(self) -> EvalueParameters: ...
    def __str__(self) -> str: ...
    def __eq__(self, other: object) -> bool: ...
    @property
    def m_mu(self) -> typing.Optional[float]: ...
    @m_mu.setter
    def m_mu(self, m: typing.Optional[float]) -> None: ...
    @property
    def m_lambda(self) -> typing.Optional[float]: ...
    @m_lambda.setter
    def m_lambda(self, l: typing.Optional[float]) -> None: ...
    @property
    def v_mu(self) -> typing.Optional[float]: ...
    @v_mu.setter
    def v_mu(self, m: typing.Optional[float]) -> None: ...
    @property
    def v_lambda(self) -> typing.Optional[float]: ...
    @v_lambda.setter
    def v_lambda(self, l: typing.Optional[float]) -> None: ...
    @property
    def f_tau(self) -> typing.Optional[float]: ...
    @f_tau.setter
    def f_tau(self, t: typing.Optional[float]) -> None: ...
    @property
    def f_lambda(self) -> typing.Optional[float]: ...
    @f_lambda.setter
    def f_lambda(self, l: typing.Optional[float]) -> None: ...
    def as_vector(self) -> VectorF: ...

class Hit(object):
    hits: TopHits[Any]
    def __getstate__(self) -> typing.Dict[str, object]: ...
    @property
    def name(self) -> bytes: ...
    @name.setter
    def name(self, name: bytes) -> None: ...
    @property
    def accession(self) -> typing.Optional[bytes]: ...
    @accession.setter
    def accession(self, accession: typing.Optional[bytes]) -> None: ...
    @property
    def description(self) -> typing.Optional[bytes]: ...
    @description.setter
    def description(self, description: typing.Optional[bytes]) -> None: ...
    @property
    def length(self) -> int: ...
    @property
    def score(self) -> float: ...
    @property
    def pre_score(self) -> float: ...
    @property
    def sum_score(self) -> float: ...
    @property
    def bias(self) -> float: ...
    @property
    def evalue(self) -> float: ...
    @property
    def pvalue(self) -> float: ...
    @property
    def domains(self) -> Domains: ...
    @property
    def best_domain(self) -> Domain: ...
    @property
    def included(self) -> bool: ...
    @included.setter
    def included(self, included: bool) -> None: ...
    @property
    def reported(self) -> bool: ...
    @reported.setter
    def reported(self, reported: bool) -> None: ...
    @property
    def new(self) -> bool: ...
    @new.setter
    def new(self, new: bool) -> None: ...
    @property
    def dropped(self) -> bool: ...
    @dropped.setter
    def dropped(self, dropped: bool) -> None: ...
    @property
    def duplicate(self) -> bool: ...
    @duplicate.setter
    def duplicate(self, duplicate: bool) -> None: ...

class HMM(object):
    alphabet: Alphabet
    @classmethod
    def sample(
        cls,
        alphabet: Alphabet,
        M: int,
        randomness: typing.Union[Randomness, int, None] = None,
        ungapped: bool = False,
        enumerate: bool = False,
    ) -> HMM: ...
    def __init__(self, alphabet: Alphabet, M: int, name: bytes) -> None: ...
    def __eq__(self, other: object) -> bool: ...
    def __copy__(self) -> HMM: ...
    def __deepcopy__(self, memo: typing.Dict[int, object]) -> HMM: ...
    def __sizeof__(self) -> int: ...
    def __getstate__(self) -> typing.Dict[str, object]: ...
    def __setstate__(self, state: typing.Dict[str, object]) -> None: ...
    @property
    def M(self) -> int: ...
    @property
    def name(self) -> bytes: ...
    @name.setter
    def name(self, name: bytes) -> None: ...
    @property
    def accession(self) -> typing.Optional[bytes]: ...
    @accession.setter
    def accession(self, accession: typing.Optional[bytes]) -> None: ...
    @property
    def checksum(self) -> typing.Optional[int]: ...
    @property
    def composition(self) -> typing.Optional[VectorF]: ...
    @property
    def description(self) -> typing.Optional[bytes]: ...
    @description.setter
    def description(self, description: typing.Optional[bytes]) -> None: ...
    @property
    def consensus(self) -> typing.Optional[str]: ...
    @property
    def consensus_structure(self) -> typing.Optional[str]: ...
    @property
    def consensus_accessibility(self) -> typing.Optional[str]: ...
    @property
    def command_line(self) -> typing.Optional[str]: ...
    @command_line.setter
    def command_line(self, cli: typing.Optional[str]) -> None: ...
    @property
    def nseq(self) -> typing.Optional[int]: ...
    @property
    def nseq_effective(self) -> typing.Optional[int]: ...
    @property
    def creation_time(self) -> typing.Optional[datetime.datetime]: ...
    @creation_time.setter
    def creation_time(self, ctime: typing.Optional[datetime.datetime]) -> None: ...
    @property
    def evalue_parameters(self) -> EvalueParameters: ...
    @property
    def cutoffs(self) -> Cutoffs: ...
    def copy(self) -> HMM: ...
    def match_occupancy(self) -> VectorF: ...
    def mean_match_entropy(self) -> float: ...
    def mean_match_information(self, background: Background) -> float: ...
    def mean_match_relative_entropy(self, background: Background) -> float: ...
    def renormalize(self) -> None: ...
    def scale(self, scale: float, exponential: bool = False) -> None: ...
    def set_composition(self) -> None: ...
    def set_consensus(self, sequence: typing.Optional[DigitalSequence]) -> None: ...
    def to_profile(
        self,
        background: typing.Optional[Background] = None,
        L: int = 400,
        multihit: bool = True,
        local: bool = True,
    ) -> Profile: ...
    def validate(self, tolerance: float = ...) -> None: ...
    def write(self, fh: typing.BinaryIO, binary: bool = False) -> None: ...
    def zero(self) -> None: ...

class HMMFile(typing.ContextManager[HMMFile], typing.Iterator[HMM]):
    _FORMATS: typing.ClassVar[typing.Dict[str, int]]
    _MAGIC: typing.ClassVar[typing.Dict[int, int]]
    def __init__(
        self,
        file: typing.Union[typing.AnyStr, os.PathLike[typing.AnyStr], typing.BinaryIO],
        db: bool = True,
    ) -> None: ...
    def __enter__(self) -> HMMFile: ...
    def __exit__(
        self,
        exc_type: typing.Optional[typing.Type[BaseException]],
        exc_value: typing.Optional[BaseException],
        traceback: typing.Optional[types.TracebackType],
    ) -> bool: ...
    def __iter__(self) -> HMMFile: ...
    def __next__(self) -> HMM: ...
    def __repr__(self) -> str: ...
    @property
    def closed(self) -> bool: ...
    @property
    def name(self) -> typing.Optional[str]: ...
    def close(self) -> None: ...
    def is_pressed(self) -> bool: ...
    def read(self) -> typing.Optional[HMM]: ...
    def rewind(self) -> None: ...
    def optimized_profiles(self) -> HMMPressedFile: ...

class HMMPressedFile(typing.Iterator[OptimizedProfile]):
    def __init__(
        self,
        file: typing.Union[typing.AnyStr, os.PathLike[typing.AnyStr]],
    ) -> None: ...
    def __enter__(self) -> HMMPressedFile: ...
    def __exit__(
        self,
        exc_type: typing.Optional[typing.Type[BaseException]],
        exc_value: typing.Optional[BaseException],
        traceback: typing.Optional[types.TracebackType],
    ) -> bool: ...
    def __iter__(self) -> HMMPressedFile: ...
    def __next__(self) -> OptimizedProfile: ...
    def __len__(self) -> int: ...
    @property
    def closed(self) -> bool: ...
    @property
    def name(self) -> str: ...
    def close(self) -> None: ...
    def read(self) -> typing.Optional[OptimizedProfile]: ...
    def rewind(self) -> None: ...

class IterationResult(typing.NamedTuple):
    hmm: HMM
    hits: TopHits[HMM]
    msa: DigitalMSA
    converged: bool
    iteration: int

class IterativeSearch(typing.Iterator[IterationResult]):
    pipeline: Pipeline
    builder: Builder
    query: typing.Union[DigitalSequence, HMM]
    converged: bool
    targets: DigitalSequenceBlock
    ranking: KeyHash
    iteration: int
    def __init__(
        self,
        pipeline: Pipeline,
        builder: Builder,
        query: typing.Union[DigitalSequence, HMM],
        targets: DigitalSequenceBlock,
        select_hits: typing.Optional[typing.Callable[[TopHits[HMM]], None]] = None,
    ) -> None: ...
    def __iter__(self) -> IterativeSearch: ...
    def __next__(self) -> IterationResult: ...

class OptimizedProfile(object):
    alphabet: Alphabet
    def __init__(self, M: int, abc: Alphabet) -> None: ...
    def __copy__(self) -> OptimizedProfile: ...
    def __deepcopy__(self, memo: typing.Dict[int, object]) -> OptimizedProfile: ...
    def __eq__(self, other: object) -> bool: ...
    def __sizeof__(self) -> int: ...
    @property
    def M(self) -> int: ...
    @property
    def L(self) -> int: ...
    @L.setter
    def L(self, L: int) -> None: ...
    @property
    def name(self) -> typing.Optional[bytes]: ...
    @property
    def accession(self) -> typing.Optional[bytes]: ...
    @property
    def description(self) -> typing.Optional[bytes]: ...
    @property
    def consensus(self) -> typing.Optional[str]: ...
    @property
    def consensus_structure(self) -> typing.Optional[str]: ...
    @property
    def reference(self) -> typing.Optional[str]: ...
    @property
    def model_mask(self) -> typing.Optional[str]: ...
    # --- MSV Filter ---
    @property
    def rbv(self) -> MatrixU8: ...
    @property
    def sbv(self) -> MatrixU8: ...
    @property
    def tbm(self) -> int: ...
    @property
    def tec(self) -> int: ...
    @property
    def tjb(self) -> int: ...
    @property
    def scale_b(self) -> float: ...
    @property
    def base_b(self) -> int: ...
    @property
    def bias_b(self) -> int: ...
    # --- ViterbiFilter ---
    @property
    def scale_w(self) -> float: ...
    @property
    def base_w(self) -> int: ...
    @property
    def ddbound_w(self) -> int: ...
    @property
    def ncj_roundoff(self) -> float: ...
    # --- Forward, Backard ---
    @property
    def rfv(self) -> MatrixF: ...
    @property
    def tfv(self) -> VectorF: ...
    @property
    def xf(self) -> MatrixF: ...
    # --- Miscellaneous ---
    @property
    def offsets(self) -> Offsets: ...
    @property
    def evalue_parameters(self) -> EvalueParameters: ...
    @property
    def cutoffs(self) -> Cutoffs: ...
    @property
    def compositions(self) -> VectorF: ...
    @property
    def local(self) -> bool: ...
    @property
    def multihit(self) -> bool: ...
    @multihit.setter
    def multihit(self, multihit: bool) -> None: ...
    def copy(self) -> OptimizedProfile: ...
    def write(
        self, fh_filter: typing.BinaryIO, fh_profile: typing.BinaryIO
    ) -> None: ...
    def convert(self, profile: Profile) -> None: ...
    def ssv_filter(self, seq: DigitalSequence) -> typing.Optional[float]: ...

class OptimizedProfileBlock(typing.MutableSequence[OptimizedProfile]):
    alphabet: Alphabet
    def __init__(
        self, alphabet: Alphabet, iterable: typing.Iterable[OptimizedProfile] = ()
    ) -> None: ...
    def __len__(self) -> int: ...
    @typing.overload
    def __getitem__(self, index: int) -> OptimizedProfile: ...
    @typing.overload
    def __getitem__(self, index: slice) -> OptimizedProfileBlock: ...
    @typing.overload
    def __setitem__(self, index: int, optimized_profile: OptimizedProfile) -> None: ...
    @typing.overload
    def __setitem__(
        self, index: slice, optimized_profile: typing.Iterable[OptimizedProfile]
    ) -> None: ...
    @typing.overload
    def __delitem__(self, index: int) -> None: ...
    @typing.overload
    def __delitem__(self, index: slice) -> None: ...
    def __repr__(self) -> str: ...
    def __contains__(self, item: object) -> bool: ...
    def __eq__(self, other: object) -> bool: ...
    def __copy__(self) -> OptimizedProfileBlock: ...
    def copy(self) -> OptimizedProfileBlock: ...
    def clear(self) -> None: ...
    def append(self, optimized_profile: OptimizedProfile) -> None: ...
    def extend(self, iterable: typing.Iterable[OptimizedProfile]) -> None: ...
    def pop(self, index: int = -1) -> OptimizedProfile: ...
    def insert(self, index: int, optimized_profile: OptimizedProfile) -> None: ...
    def index(
        self,
        optimized_profile: OptimizedProfile,
        start: int = 0,
        end: int = sys.maxsize,
    ) -> int: ...
    def remove(self, optimized_profile: OptimizedProfile) -> None: ...

class Offsets(object):
    def __init__(self, owner: typing.Union[Profile, OptimizedProfile]) -> None: ...
    def __copy__(self) -> Offsets: ...
    def __repr__(self) -> str: ...
    @property
    def model(self) -> typing.Optional[int]: ...
    @model.setter
    def model(self, model: typing.Optional[int]) -> None: ...
    @property
    def filter(self) -> typing.Optional[int]: ...
    @filter.setter
    def filter(self, filter: typing.Optional[int]) -> None: ...
    @property
    def profile(self) -> typing.Optional[int]: ...
    @profile.setter
    def profile(self, profile: typing.Optional[int]) -> None: ...

class Pipeline(object):
    M_HINT: typing.ClassVar[int] = 100
    L_HINT: typing.ClassVar[int] = 100
    _BIT_CUTOFFS: typing.ClassVar[typing.Dict[str, int]]
    alphabet: Alphabet
    background: Background
    profile: typing.Optional[Profile]
    randomness: Randomness
    def __init__(
        self,
        alphabet: Alphabet,
        background: typing.Optional[Background] = None,
        *,
        bias_filter: bool = True,
        null2: bool = True,
        seed: int = 42,
        Z: typing.Optional[float] = None,
        domZ: typing.Optional[float] = None,
        F1: float = 0.02,
        F2: float = 1e-3,
        F3: float = 1e-5,
        E: float = 10.0,
        T: typing.Optional[float] = None,
        domE: float = 10.0,
        domT: typing.Optional[float] = None,
        incE: float = 0.01,
        incT: typing.Optional[float] = None,
        incdomE: float = 0.01,
        incdomT: typing.Optional[float] = None,
        bit_cutoffs: typing.Optional[BIT_CUTOFFS] = None,
    ) -> None: ...
    @property
    def Z(self) -> typing.Optional[float]: ...
    @Z.setter
    def Z(self, Z: typing.Optional[float]) -> None: ...
    @property
    def domZ(self) -> typing.Optional[float]: ...
    @domZ.setter
    def domZ(self, domZ: typing.Optional[float]) -> None: ...
    @property
    def null2(self) -> bool: ...
    @null2.setter
    def null2(self, null2: bool) -> None: ...
    @property
    def bias_filter(self) -> bool: ...
    @bias_filter.setter
    def bias_filter(self, bias_filter: bool) -> None: ...
    @property
    def F1(self) -> float: ...
    @F1.setter
    def F1(self, F1: float) -> None: ...
    @property
    def F2(self) -> float: ...
    @F2.setter
    def F2(self, F2: float) -> None: ...
    @property
    def F3(self) -> float: ...
    @F3.setter
    def F3(self, F3: float) -> None: ...
    @property
    def E(self) -> float: ...
    @E.setter
    def E(self, E: float) -> None: ...
    @property
    def T(self) -> typing.Optional[float]: ...
    @T.setter
    def T(self, T: typing.Optional[float]) -> None: ...
    @property
    def domE(self) -> float: ...
    @domE.setter
    def domE(self, domE: float) -> None: ...
    @property
    def domT(self) -> typing.Optional[float]: ...
    @domT.setter
    def domT(self, domT: typing.Optional[float]) -> None: ...
    @property
    def incE(self) -> float: ...
    @incE.setter
    def incE(self, incE: float) -> None: ...
    @property
    def incT(self) -> typing.Optional[float]: ...
    @incT.setter
    def incT(self, incT: typing.Optional[float]) -> None: ...
    @property
    def incdomE(self) -> float: ...
    @incdomE.setter
    def incdomE(self, incdomE: float) -> None: ...
    @property
    def incdomT(self) -> typing.Optional[float]: ...
    @incdomT.setter
    def incdomT(self, incdomT: typing.Optional[float]) -> None: ...
    @property
    def bit_cutoffs(self) -> typing.Optional[BIT_CUTOFFS]: ...
    @bit_cutoffs.setter
    def bit_cutoffs(self, bit_cutoffs: typing.Optional[BIT_CUTOFFS]) -> None: ...
    def arguments(self) -> typing.List[str]: ...
    def clear(self) -> None: ...
    def search_hmm(
        self,
        query: typing.Union[HMM, Profile, OptimizedProfile],
        sequences: typing.Union[DigitalSequenceBlock, SequenceFile[DigitalSequence]],
    ) -> TopHits[typing.Union[HMM, Profile, OptimizedProfile]]: ...
    def search_msa(
        self,
        query: DigitalMSA,
        sequences: typing.Union[DigitalSequenceBlock, SequenceFile[DigitalSequence]],
        builder: typing.Optional[Builder] = None,
    ) -> TopHits[DigitalMSA]: ...
    def search_seq(
        self,
        query: DigitalSequence,
        sequences: typing.Union[DigitalSequenceBlock, SequenceFile[DigitalSequence]],
        builder: typing.Optional[Builder] = None,
    ) -> TopHits[DigitalSequence]: ...
    def scan_seq(
        self,
        query: DigitalSequence,
        optimized_profiles: typing.Union[OptimizedProfileBlock, HMMPressedFile],
    ) -> TopHits[DigitalSequence]: ...
    def iterate_seq(
        self,
        query: DigitalSequence,
        sequences: DigitalSequenceBlock,
        builder: typing.Optional[Builder] = None,
        select_hits: typing.Optional[typing.Callable[[TopHits[DigitalSequence]], None]] = None,
    ) -> IterativeSearch: ...
    def iterate_hmm(
        self,
        query: HMM,
        sequences: DigitalSequenceBlock,
        builder: typing.Optional[Builder] = None,
        select_hits: typing.Optional[typing.Callable[[TopHits[DigitalSequence]], None]] = None,
    ) -> IterativeSearch: ...

class LongTargetsPipeline(Pipeline):
    def __init__(
        self,
        alphabet: Alphabet,
        background: typing.Optional[Background] = None,
        *,
        strand: typing.Optional[STRAND] = None,
        B1: int = 100,
        B2: int = 240,
        B3: int = 1000,
        block_length: int = 1024 * 256,
        window_length: typing.Optional[int] = None,
        window_beta: typing.Optional[float] = None,
        bias_filter: bool = True,
        null2: bool = True,
        seed: typing.Optional[int] = None,
        Z: typing.Optional[float] = None,
        domZ: typing.Optional[float] = None,
        F1: float = 0.02,
        F2: float = 1e-3,
        F3: float = 1e-5,
        E: float = 10.0,
        T: typing.Optional[float] = None,
        domE: float = 10.0,
        domT: typing.Optional[float] = None,
        incE: float = 0.01,
        incT: typing.Optional[float] = None,
        incdomE: float = 0.01,
        incdomT: typing.Optional[float] = None,
        bit_cutoffs: typing.Optional[BIT_CUTOFFS] = None,
    ) -> None: ...
    @property
    def B1(self) -> int: ...
    @B1.setter
    def B1(self, B1: int) -> None: ...
    @property
    def B2(self) -> int: ...
    @B2.setter
    def B2(self, B2: int) -> None: ...
    @property
    def B3(self) -> int: ...
    @B3.setter
    def B3(self, B3: int) -> None: ...
    @property
    def strand(self) -> typing.Optional[STRAND]: ...
    @strand.setter
    def strand(self, strand: typing.Optional[STRAND]) -> None: ...
    @property
    def window_beta(self) -> float: ...
    @window_beta.setter
    def window_beta(self, window_beta: float) -> None: ...
    @property
    def window_length(self) -> typing.Optional[int]: ...
    @window_length.setter
    def window_length(self, window_length: typing.Optional[int]) -> None: ...

class Profile(object):
    alphabet: Alphabet
    def __init__(self, M: int, alphabet: Alphabet) -> None: ...
    def __copy__(self) -> Profile: ...
    def __deepcopy__(self, memo: typing.Dict[int, object]) -> Profile: ...
    def __eq__(self, other: object) -> bool: ...
    def __sizeof__(self) -> int: ...
    @property
    def M(self) -> int: ...
    @property
    def name(self) -> typing.Optional[bytes]: ...
    @property
    def accession(self) -> typing.Optional[bytes]: ...
    @property
    def description(self) -> typing.Optional[bytes]: ...
    @property
    def consensus(self) -> typing.Optional[str]: ...
    @property
    def consensus_structure(self) -> typing.Optional[str]: ...
    @property
    def cutoffs(self) -> Cutoffs: ...
    @property
    def evalue_parameters(self) -> EvalueParameters: ...
    @property
    def offsets(self) -> Offsets: ...
    @property
    def local(self) -> bool: ...
    @property
    def multihit(self) -> bool: ...
    @multihit.setter
    def multihit(self, multihit: bool) -> None: ...
    def clear(self) -> None: ...
    def configure(
        self,
        hmm: HMM,
        background: Background,
        L: int = 400,
        multihit: bool = True,
        local: bool = True,
    ) -> None: ...
    def copy(self) -> Profile: ...
    def to_optimized(self) -> OptimizedProfile: ...

class ScoreData(object):
    Kp: int
    def __init__(self, gm: Profile, om: OptimizedProfile) -> None: ...
    def __copy__(self) -> ScoreData: ...
    def copy(self) -> ScoreData: ...

class TopHits(typing.Sequence[Hit], typing.Generic[Q]):
    def __init__(self, query: Q) -> None: ...
    def __bool__(self) -> bool: ...
    def __copy__(self) -> TopHits[Q]: ...
    def __deepcopy__(self, memo: typing.Dict[int, object]) -> TopHits[Q]: ...
    def __len__(self) -> int: ...
    @typing.overload
    def __getitem__(self, index: int) -> Hit: ...
    @typing.overload
    def __getitem__(self, index: slice) -> typing.Sequence[Hit]: ...
    def __iadd__(self, other: TopHits[Q]) -> TopHits[Q]: ...
    def __getstate__(self) -> typing.Dict[str, object]: ...
    def __setstate__(self, state: typing.Dict[str, object]) -> None: ...
    @property
    def mode(self) -> HITS_MODE: ...
    @property
    def query(self) -> Q: ...
    @property
    def Z(self) -> float: ...
    @property
    def domZ(self) -> float: ...
    @property
    def long_targets(self) -> bool: ...
    @property
    def E(self) -> float: ...
    @property
    def T(self) -> typing.Optional[float]: ...
    @property
    def domE(self) -> float: ...
    @property
    def domT(self) -> typing.Optional[float]: ...
    @property
    def incE(self) -> float: ...
    @property
    def incT(self) -> typing.Optional[float]: ...
    @property
    def incdomE(self) -> float: ...
    @property
    def incdomT(self) -> typing.Optional[float]: ...
    @property
    def bit_cutoffs(self) -> typing.Optional[BIT_CUTOFFS]: ...
    @property
    def searched_models(self) -> int: ...
    @property
    def searched_nodes(self) -> int: ...
    @property
    def searched_sequences(self) -> int: ...
    @property
    def searched_residues(self) -> int: ...
    @property
    def strand(self) -> typing.Optional[STRAND]: ...
    @property
    def block_length(self) -> typing.Optional[int]: ...
    @property
    def reported(self) -> SizedIterator[Hit]: ...
    @property
    def included(self) -> SizedIterator[Hit]: ...
    def compare_ranking(self, ranking: KeyHash) -> int: ...
    def sort(self, by: SORT_KEY = "key") -> None: ...
    def is_sorted(self, by: SORT_KEY = "key") -> bool: ...
    def copy(self) -> TopHits[Q]: ...
    def merge(self, *others: TopHits[Q]) -> TopHits[Q]: ...
    def to_msa(
        self,
        alphabet: Alphabet,
        sequences: typing.Optional[typing.List[Sequence]] = None,
        traces: typing.Optional[typing.List[Trace]] = None,
        *,
        trim: bool = False,
        digitize: bool = False,
        all_consensus_cols: bool = False,
    ) -> MSA: ...
    def write(
        self,
        fh: typing.BinaryIO,
        format: HITS_FORMAT = "targets",
        header: bool = True,
    ) -> None: ...

class Trace(object):
    @classmethod
    def from_sequence(cls, sequence: Sequence) -> Trace: ...
    def __init__(self, posteriors: bool = False) -> None: ...
    def __eq__(self, other: object) -> bool: ...
    @property
    def M(self) -> int: ...
    @property
    def L(self) -> int: ...
    @property
    def posterior_probabilities(self) -> typing.Optional[VectorF]: ...
    def expected_accuracy(self) -> float: ...
    def score(self, sequence: DigitalSequence, profile: Profile) -> float: ...

class Traces(typing.MutableSequence[Trace]):
    def __init__(self, iterable: Iterable[Trace] = None) -> None: ...
    def __eq__(self, other: object) -> bool: ...
    def __len__(self) -> int: ...
    @typing.overload
    def __getitem__(self, index: int) -> Trace: ...
    @typing.overload
    def __getitem__(self, index: slice) -> typing.Sequence[Trace]: ...
    @typing.overload
    def __setitem__(self, index: int, item: Trace) -> None: ...
    @typing.overload
    def __setitem__(self, index: slice, item: Trace) -> None: ...
    def __contains__(self, other: object) -> bool: ...
    def append(self, trace: Trace) -> None: ...
    def clear(self) -> None: ...
    def extend(self, iterable: Iterable[Trace]) -> None: ...
    def pop(self, index: int = -1) -> None: ...
    def insert(self, index: int, trace: Trace) -> None: ...
    def index(self, trace: Trace, start: int = 0, stop: int = sys.maxsize) -> int: ...
    def remove(self, trace: Trace) -> None: ...

class TraceAligner(object):
    def __init__(self) -> None: ...
    def compute_traces(
        self,
        hmm: HMM,
        sequences: DigitalSequenceBlock,
    ) -> Traces: ...
    @typing.overload
    def align_traces(
        self,
        hmm: HMM,
        sequences: DigitalSequenceBlock,
        traces: Traces,
        digitize: Literal[True],
        trim: bool = False,
        all_consensus_cols: bool = False,
    ) -> DigitalMSA: ...
    @typing.overload
    def align_traces(
        self,
        hmm: HMM,
        sequences: DigitalSequenceBlock,
        traces: Traces,
        digitize: Literal[False],
        trim: bool = False,
        all_consensus_cols: bool = False,
    ) -> TextMSA: ...
    @typing.overload
    def align_traces(
        self,
        hmm: HMM,
        sequences: DigitalSequenceBlock,
        traces: Traces,
        digitize: bool = False,
        trim: bool = False,
        all_consensus_cols: bool = False,
    ) -> MSA: ...

class Transitions(enum.IntEnum):
    MM = ...
    MI = ...
    MD = ...
    IM = ...
    II = ...
    DM = ...
    DD = ...



cdef extern from "esl_distance.h" nogil:

    # Pairwise distances for aligned text sequences.
    int esl_dst_CPairId(const char *asq1, const char *asq2,
    			   double *opt_pid, int *opt_nid, int *opt_n);
    int esl_dst_CPairmatch(const char *asq1, const char *asq2,
    			      double *opt_pmatch, int *opt_nmatch, int *opt_n);
    int esl_dst_CJukesCantor(int K, const char *as1, const char *as2,
    				double *opt_distance, double *opt_variance);

    # Pairwise distances for aligned digital seqs.
    int esl_dst_XPairId(const ESL_ALPHABET *abc, const ESL_DSQ *ax1, const ESL_DSQ *ax2,
    			   double *opt_pid, int *opt_nid, int *opt_n);
    int esl_dst_XPairMatch(const ESL_ALPHABET *abc, const ESL_DSQ *ax1, const ESL_DSQ *ax2,
    			      double *opt_distance, int *opt_nmatch, int *opt_n);
    int esl_dst_XJukesCantor(const ESL_ALPHABET *abc, const ESL_DSQ *ax, const ESL_DSQ *ay,
    				double *opt_distance, double *opt_variance);

    # Distance matrices for aligned text sequences.
    int esl_dst_CPairIdMx     (char **as, int N, ESL_DMATRIX **ret_S);
    int esl_dst_CDiffMx       (char **as, int N, ESL_DMATRIX **ret_D);
    int esl_dst_CJukesCantorMx(int K, char **as, int N, ESL_DMATRIX **opt_D, ESL_DMATRIX **opt_V);

    # Distance matrices for aligned digital sequences.
    int esl_dst_XPairIdMx(const ESL_ALPHABET *abc, ESL_DSQ **ax, int N, ESL_DMATRIX **ret_S);
    int esl_dst_XDiffMx  (const ESL_ALPHABET *abc, ESL_DSQ **ax, int N, ESL_DMATRIX **ret_D);

    int esl_dst_XJukesCantorMx(const ESL_ALPHABET *abc, ESL_DSQ **ax, int nseq,
    				  ESL_DMATRIX **opt_D, ESL_DMATRIX **opt_V);

    # Average pairwise identity for multiple alignments.
    int esl_dst_CAverageId   (char **as, int nseq, int max_comparisons, double *ret_id);
    int esl_dst_CAverageMatch(char **as, int N, int max_comparisons, double *ret_match);

    int esl_dst_XAverageId   (const ESL_ALPHABET *abc, ESL_DSQ **ax, int N, int max_comparisons, double *ret_id);
    int esl_dst_XAverageMatch(const ESL_ALPHABET *abc, ESL_DSQ **ax, int N, int max_comparisons, double *ret_match);



from libc.stdint cimport int32_t, int64_t, uint32_t
from posix.types cimport off_t

from libeasel cimport ESL_DSQ
from libeasel.alphabet cimport ESL_ALPHABET
from libeasel.msa cimport ESL_MSA
from libeasel.random cimport ESL_RANDOMNESS


cdef extern from "esl_sq.h" nogil:

    ctypedef struct ESL_SQ:
        char* name
        char* acc
        char* desc
        int32_t tax_id
        char* seq
        ESL_DSQ* dsq
        char* ss
        int64_t n

        int64_t start
        int64_t end
        int64_t C
        int64_t W
        int64_t L

        char* source
        int nalloc
        int aalloc
        int dalloc
        int64_t salloc
        int srcalloc

        int64_t idx
        off_t roff
        off_t hoff
        off_t doff
        off_t eoff

        char** xr_tag
        char** xr
        int nxr

        ESL_ALPHABET* abc


    ctypedef struct ESL_SQ_BLOCK:
        int count
        int listSize
        int complete
        int64_t first_seqidx
        ESL_SQ* list


    ESL_SQ *esl_sq_Create()
    ESL_SQ *esl_sq_CreateFrom(const char *name, const char *seq,
    				 const char *desc, const char *acc, const char *ss)
    int     esl_sq_Grow  (ESL_SQ *sq, int64_t *ret_nsafe)
    int     esl_sq_GrowTo(ESL_SQ *sq, int64_t  n)
    int     esl_sq_Copy(const ESL_SQ *src, ESL_SQ *dst)
    int     esl_sq_Compare  (ESL_SQ *sq1, ESL_SQ *sq2)
    int     esl_sq_Reuse    (ESL_SQ *sq)
    int     esl_sq_IsDigital(const ESL_SQ *sq)
    int     esl_sq_IsText   (const ESL_SQ *sq)
    void    esl_sq_Destroy  (ESL_SQ *sq)

    int     esl_sq_SetName        (ESL_SQ *sq, const char *name)
    int     esl_sq_SetAccession   (ESL_SQ *sq, const char *acc)
    int     esl_sq_SetDesc        (ESL_SQ *sq, const char *desc)
    int     esl_sq_SetSource      (ESL_SQ *sq, const char *source)
    int     esl_sq_FormatName     (ESL_SQ *sq, const char *name,   ...) #ESL_ATTRIBUTE_FORMAT(printf, 2, 3)
    int     esl_sq_FormatAccession(ESL_SQ *sq, const char *acc,    ...) #ESL_ATTRIBUTE_FORMAT(printf, 2, 3)
    int     esl_sq_FormatDesc     (ESL_SQ *sq, const char *desc,   ...) #ESL_ATTRIBUTE_FORMAT(printf, 2, 3)
    int     esl_sq_FormatSource   (ESL_SQ *sq, const char *source, ...) #ESL_ATTRIBUTE_FORMAT(printf, 2, 3)
    int     esl_sq_AppendDesc     (ESL_SQ *sq, const char *desc)
    int     esl_sq_SetCoordComplete(ESL_SQ *sq, int64_t L)
    int     esl_sq_CAddResidue (ESL_SQ *sq, char c)
    int     esl_sq_ReverseComplement(ESL_SQ *sq)
    int     esl_sq_Checksum(const ESL_SQ *sq, uint32_t *ret_checksum)
    int     esl_sq_CountResidues(const ESL_SQ *sq, int start, int L, float *f)

    ESL_SQ *esl_sq_CreateDigital(const ESL_ALPHABET *abc)
    ESL_SQ *esl_sq_CreateDigitalFrom(const ESL_ALPHABET *abc, const char *name, const ESL_DSQ *dsq,
    					int64_t L, const char *desc, const char *acc,  const char *ss);
    int     esl_sq_Digitize(const ESL_ALPHABET *abc, ESL_SQ *sq)
    int     esl_sq_Textize(ESL_SQ *sq)
    int     esl_sq_GuessAlphabet(ESL_SQ *sq, int *ret_type)
    int     esl_sq_XAddResidue(ESL_SQ *sq, ESL_DSQ x)
    int     esl_sq_ConvertDegen2X(ESL_SQ *sq)

    int     esl_sq_GetFromMSA  (const ESL_MSA *msa, int which, ESL_SQ *sq)
    int     esl_sq_FetchFromMSA(const ESL_MSA *msa, int which, ESL_SQ **ret_sq)

    ESL_SQ_BLOCK *esl_sq_CreateBlock(int count)
    int esl_sq_BlockGrowTo(ESL_SQ_BLOCK *sqblock, int newsize, int do_digital, const ESL_ALPHABET *abc)
    ESL_SQ_BLOCK *esl_sq_CreateDigitalBlock(int count, const ESL_ALPHABET *abc)
    void          esl_sq_DestroyBlock(ESL_SQ_BLOCK *sqBlock)
    int esl_sq_BlockReallocSequences(ESL_SQ_BLOCK *block)
    int esl_sq_Sample(ESL_RANDOMNESS *rng, ESL_ALPHABET *abc, int maxL, ESL_SQ **ret_sq)



# coding: utf-8
"""Cython bindings and Python interface to HMMER3.

HMMER is a biological sequence analysis tool that uses profile hidden Markov
models to search for sequence homologs. HMMER3 is maintained by members of the
the `Eddy/Rivas Laboratory <http://eddylab.org/>`_ at Harvard University.

``pyhmmer`` is a module, implemented using the `Cython <https://cython.org/>`_
language, that provides bindings to HMMER3. It directly interacts with the
HMMER internals, which has several advantages over CLI wrappers like
`hmmer-py <https://pypi.org/project/hmmer/>`_.

References:
    Larralde, Martin, and Georg Zeller.
    "PyHMMER: A Python Library Binding to HMMER for Efficient
    Sequence Analysis". Bioinformatics, 19 April 2023, btad214.
    :doi:`10.1093/bioinformatics/btad214`. :pmid:`37074928`.

"""
import collections.abc as _collections_abc
import contextlib as _contextlib
import os as _os

from . import errors
from . import easel
from . import plan7
from . import daemon
from ._version import __version__
from .hmmer import hmmalign, hmmsearch, hmmpress, nhmmer, hmmscan, phmmer, jackhmmer


__author__ = "Martin Larralde <martin.larralde@embl.de>"
__license__ = "MIT"
__all__ = [
    "errors",
    "easel",
    "plan7",
    "daemon",
    "hmmalign",
    "hmmsearch",
    "hmmscan",
    "hmmpress",
    "phmmer",
    "nhmmer",
    "jackhmmer",
]

# Small addition to the docstring: we want to show a link redirecting to the
# rendered version of the documentation, but this can only work when Python
# is running with docstrings enabled
if __doc__ is not None:
    __doc__ += """See Also:
    An online rendered version of the documentation for this version of the
    library on `Read The Docs <https://pyhmmer.readthedocs.io/en/v{}/>`_.

    """.format(__version__)

# Register collections using the `collections.abc` module (this is probably
# not required with later versions of Python)
_collections_abc.Iterator.register(easel.SequenceFile)
_collections_abc.Iterator.register(plan7.HMMFile)
_collections_abc.Mapping.register(easel.KeyHash)
_collections_abc.Sized.register(plan7.Alignment)
_collections_abc.Sequence.register(easel.Bitfield)
_collections_abc.Sequence.register(plan7.Domains)
_collections_abc.Sequence.register(plan7.TopHits)
_collections_abc.Sequence.register(easel.SequenceBlock)
_collections_abc.Sequence.register(plan7.OptimizedProfileBlock)

if hasattr(_contextlib, "AbstractContextManager"):
    _contextlib.AbstractContextManager.register(easel.SequenceFile)
    _contextlib.AbstractContextManager.register(easel.SSIReader)
    _contextlib.AbstractContextManager.register(easel.SSIWriter)
    _contextlib.AbstractContextManager.register(plan7.HMMFile)



import os
import typing
import functools
import itertools
from multiprocessing.pool import ThreadPool

import psutil

from ..easel import DigitalSequence, DigitalSequenceBlock, MSA
from ..plan7 import HMM, TraceAligner, Traces

if typing.TYPE_CHECKING:
    from ._base import BACKEND

# --- Utils ------------------------------------------------------------------

def _batch_sequences(sequences: DigitalSequenceBlock, cpus: int) -> typing.Iterable[DigitalSequenceBlock]:
    batchsize = (len(sequences) + cpus - 1) // cpus
    indices = [0]
    for i in range(cpus):
        indices.append(indices[i] + batchsize)
    indices[-1] = len(sequences)
    return (sequences[indices[i]:indices[i+1]] for i in range(cpus))

# --- hmmalign ---------------------------------------------------------------

def hmmalign(
    hmm: HMM,
    sequences: typing.Iterable[DigitalSequence],
    *,
    cpus: int = 0,
    digitize: bool = False,
    trim: bool = False,
    all_consensus_cols: bool = True,
) -> MSA:
    """Align several sequences to a reference HMM, and return the MSA.

    Arguments:
        hmm (`~pyhmmer.plan7.HMM`): The reference HMM to use for the
            alignment.
        sequences (iterable of `~pyhmmer.easel.DigitalSequence`): The
            sequences to align to the HMM. If you plan on using the
            same sequences several times, consider storing them into
            a `~pyhmmer.easel.DigitalSequenceBlock` directly.

    Keyword Arguments:
        cpus (`int`): The number of threads to run in parallel. Pass ``1``
            to run everything in the main thread, ``0`` to automatically
            select a suitable number (using `psutil.cpu_count`), or any
            positive number otherwise.
        trim (`bool`): Trim off any residues that get assigned to
            flanking :math:`N` and :math:`C` states (in profile traces)
            or :math:`I_0` and :math:`I_m` (in core traces).
        digitize (`bool`): If set to `True`, returns a `DigitalMSA`
            instead of a `TextMSA`.
        all_consensus_cols (`bool`): Force a column to be created for
            every consensus column in the model, even if it means having
            all gap character in a column.

    Returns:
        `~pyhmmer.easel.MSA`: A multiple sequence alignment containing
        the aligned sequences, either a `TextMSA` or a `DigitalMSA`
        depending on the value of the ``digitize`` argument.

    See Also:
        The `~pyhmmer.plan7.TraceAligner` class, which lets you inspect the
        intermediate tracebacks obtained for each alignment before building
        a MSA.

    .. versionadded:: 0.4.7

    .. versionadded:: 0.11.4
        The ``cpus`` argument for parallel processing with multithreading.

    """
    cpus = cpus if cpus > 0 else psutil.cpu_count(logical=False) or os.cpu_count() or 1

    aligner = TraceAligner()
    if not isinstance(sequences, DigitalSequenceBlock):
        sequences = DigitalSequenceBlock(hmm.alphabet, sequences)

    if len(sequences) < cpus and cpus > 1:
        cpus = len(sequences)

    if cpus == 1:
        traces = aligner.compute_traces(hmm, sequences)
    else:
        batches = _batch_sequences(sequences, cpus)
        with ThreadPool(cpus) as pool:
            traces = Traces()
            _compute_traces = functools.partial(aligner.compute_traces, hmm)
            for batch_traces in pool.map(_compute_traces, batches):
                traces.extend(batch_traces)

    return aligner.align_traces(
        hmm,
        sequences,
        traces,
        trim=trim,
        digitize=digitize,
        all_consensus_cols=all_consensus_cols,
    )




import collections
import copy
import os
import queue
import multiprocessing
import typing
import threading

import psutil

from ..easel import Alphabet, DigitalSequence, DigitalMSA, DigitalSequenceBlock, SequenceFile
from ..plan7 import TopHits, HMM, Profile, OptimizedProfile, LongTargetsPipeline, Builder
from ..utils import singledispatchmethod, peekable
from ._base import _BaseDispatcher, _BaseWorker, _BaseChore, _AnyProfile

_NHMMERQueryType = typing.Union[DigitalSequence, DigitalMSA, _AnyProfile]
_N = typing.TypeVar("_N", DigitalSequence, DigitalMSA, HMM, Profile, OptimizedProfile)

if typing.TYPE_CHECKING:
    from ._base import Unpack, LongTargetsPipelineOptions, BACKEND

# --- Worker -------------------------------------------------------------------

class _NHMMERWorker(
    _BaseWorker[
        _NHMMERQueryType,
        typing.Union[DigitalSequenceBlock, "SequenceFile[DigitalSequence]"],
        "TopHits[_NHMMERQueryType]",
    ],
    threading.Thread
):

    pipeline_class = LongTargetsPipeline

    @singledispatchmethod
    def query(self, query) -> "TopHits[Any]":  # type: ignore
        raise TypeError(
            "Unsupported query type for `nhmmer`: {}".format(type(query).__name__)
        )

    @query.register(DigitalSequence)
    def _(self, query: DigitalSequence) -> "TopHits[DigitalSequence]":  # type: ignore
        assert self.pipeline is not None
        return self.pipeline.search_seq(query, self.targets, self.builder)

    @query.register(DigitalMSA)
    def _(self, query: DigitalMSA) -> "TopHits[DigitalMSA]":  # type: ignore
        assert self.pipeline is not None
        return self.pipeline.search_msa(query, self.targets, self.builder)

    @query.register(HMM)
    @query.register(Profile)
    @query.register(OptimizedProfile)
    def _(self, query: _AnyProfile) -> "TopHits[_AnyProfile]":  # type: ignore
        assert self.pipeline is not None
        return self.pipeline.search_hmm(query, self.targets)


class _NHMMERThread(_NHMMERWorker, threading.Thread):
    pass


class _NHMMERProcess(_NHMMERWorker, multiprocessing.Process):
    pass


# --- Dispatcher ---------------------------------------------------------------

class _NHMMERDispatcher(
    _BaseDispatcher[
        _NHMMERQueryType,
        typing.Union[DigitalSequenceBlock, "SequenceFile[DigitalSequence]"],
        "TopHits[_NHMMERQueryType]",
    ]
):
    def __init__(
        self,
        queries: typing.Iterable[_NHMMERQueryType],
        targets: typing.Union[DigitalSequenceBlock, "SequenceFile[DigitalSequence]"],
        cpus: int = 0,
        callback: typing.Optional[
            typing.Callable[[_NHMMERQueryType, int], None]
        ] = None,
        builder: typing.Optional[Builder] = None,
        timeout: int = 1,
        **options,  # type: Unpack[LongTargetsPipelineOptions]
    ) -> None:
        super().__init__(
            queries=queries,
            targets=targets,
            cpus=cpus,
            callback=callback,
            builder=builder,
            timeout=timeout,
            **options,  # type: ignore
        )

    def _new_worker(
        self,
        query_queue: "queue.Queue[typing.Optional[_BaseChore[_NHMMERQueryType, TopHits[_NHMMERQueryType]]]]",
        query_count: "multiprocessing.Value[int]",  # type: ignore
        kill_switch: threading.Event,
    ) -> _NHMMERWorker:
        if isinstance(self.targets, SequenceFile):
            assert self.targets.name is not None
            targets = SequenceFile(
                self.targets.name,
                format=self.targets.format,
                digital=True,
                alphabet=self.options["alphabet"],
            )
        else:
            targets = self.targets  # type: ignore
        params = [
            targets,
            query_queue,
            query_count,
            kill_switch,
            self.callback,
            self.options,
            copy.copy(self.builder),
        ]
        if self.backend == "threading":
            return _NHMMERThread(*params)
        elif self.backend == "multiprocessing":
            return _NHMMERProcess(*params)
        else:
            raise ValueError(f"Invalid backend for `nhmmer`: {self.backend!r}")


# --- nhmmer -----------------------------------------------------------------

def nhmmer(
    queries: typing.Union[_N, typing.Iterable[_N]],
    sequences: typing.Iterable[DigitalSequence],
    *,
    cpus: int = 0,
    callback: typing.Optional[typing.Callable[[_N, int], None]] = None,
    backend: "BACKEND" = "threading",
    builder: typing.Optional[Builder] = None,
    **options,  # type: Unpack[LongTargetsPipelineOptions]
) -> typing.Iterator["TopHits[_N]"]:
    """Search nucleotide sequences against a sequence database.

    Arguments:
        queries (iterable of `DigitalSequence`, `DigitalMSA`, `HMM`): The
            query sequences or profiles to search for in the sequence
            database. Passing a single object is supported.
        sequences (iterable of `~pyhmmer.easel.DigitalSequence`): A
            database of sequences to query. If you plan on using the
            same sequences several times, consider storing them into
            a `~pyhmmer.easel.DigitalSequenceBlock` directly. If a
            `SequenceFile` is given, profiles will be loaded iteratively
            from disk rather than prefetched.
        cpus (`int`): The number of threads to run in parallel. Pass ``1`` to
            run everything in the main thread, ``0`` to automatically
            select a suitable number (using `psutil.cpu_count`), or any
            positive number otherwise.
        callback (callable): A callback that is called everytime a query is
            processed with two arguments: the query, and the total number
            of queries. This can be used to display progress in UI.
        builder (`~pyhmmer.plan7.Builder`, optional): A builder to configure
            how the queries are converted to HMMs. Passing `None` will create
            a default instance.
        backend (`str`): The parallel backend to use for workers to be
            executed. Supports ``threading`` to use thread-based parallelism,
            or ``multiprocessing`` to use process-based parallelism.

    Yields:
        `~pyhmmer.plan7.TopHits`: A *top hits* instance for each query,
        in the same order the queries were passed in the input.

    Note:
        Any additional keyword arguments passed to the `nhmmer` function
        will be passed to the `~pyhmmer.plan7.LongTargetsPipeline` created
        in each worker thread. The ``strand`` argument can be used to
        restrict the search on the direct or reverse strand.

    Caution:
        This function is not just `phmmer` for nucleotide sequences; it
        actually uses a `~pyhmmer.plan7.LongTargetsPipeline` internally
        instead of processing each target sequence in its entirety when
        searching for hits. This avoids hitting the maximum target size
        that can be used (100,000 residues), which may be a problem for
        some larger genomes.

    .. versionadded:: 0.3.0

    .. versionchanged:: 0.4.9
       Allow using `Profile` and `OptimizedProfile` queries.

    .. versionchanged:: 0.7.0
        Queries may now be an iterable of different types, or a single object.

    """
    cpus = cpus if cpus > 0 else psutil.cpu_count(logical=False) or os.cpu_count() or 1
    alphabet = options.get("alphabet")

    if not isinstance(queries, collections.abc.Iterable):
        queries = (queries,)

    if isinstance(sequences, SequenceFile):
        if sequences.name is None:
            raise ValueError("expected named `SequenceFile` for targets")
        if not sequences.digital:
            raise ValueError("expected digital mode `SequenceFile` for targets")
        assert sequences.alphabet is not None
        alphabet = alphabet or sequences.alphabet
        targets: typing.Union["SequenceFile[DigitalSequence]", DigitalSequenceBlock] = sequences
    elif isinstance(sequences, DigitalSequenceBlock):
        alphabet = alphabet or sequences.alphabet
        targets = sequences
    else:
        alphabet = alphabet or Alphabet.dna()
        targets = DigitalSequenceBlock(alphabet, sequences)

    if builder is None:
        builder = Builder(
            alphabet,
            seed=options.get("seed", 42),
            window_length=options.get("window_length"),
            window_beta=options.get("window_beta"),
        )

    if "alphabet" not in options:
        options["alphabet"] = alphabet
    dispatcher = _NHMMERDispatcher(
        queries=queries,
        targets=targets,
        cpus=cpus,
        backend=backend,
        callback=callback,  # type: ignore
        builder=builder,
        **options,
    )
    return dispatcher.run()  # type: ignore




cdef extern from "esl_msafile_selex.h" nogil:
    int esl_msafile_selex_SetInmap     (ESL_MSAFILE *afp)
    int esl_msafile_selex_GuessAlphabet(ESL_MSAFILE *afp, int *ret_type)
    int esl_msafile_selex_Read         (ESL_MSAFILE *afp, ESL_MSA **ret_msa)
    int esl_msafile_selex_Write        (FILE *fp,    const ESL_MSA *msa)
